@@ -3,7 +3,7 @@ package YAHC;
 use strict;
 use warnings;
 
-our $VERSION = '0.027';
+our $VERSION = '0.028';
 
 use EV;
 use Time::HiRes;
@@ -21,6 +21,7 @@ sub YAHC::Error::REQUEST_TIMEOUT         () { 1 << 0 }
 sub YAHC::Error::CONNECT_TIMEOUT         () { 1 << 1 }
 sub YAHC::Error::DRAIN_TIMEOUT           () { 1 << 2 }
 sub YAHC::Error::LIFETIME_TIMEOUT        () { 1 << 3 }
+sub YAHC::Error::TIMEOUT                 () { 1 << 8 }
 sub YAHC::Error::RETRY_LIMIT             () { 1 << 9 }
 
 sub YAHC::Error::CONNECT_ERROR           () { 1 << 10 }
@@ -68,8 +69,10 @@ our @EXPORT_OK = qw/
     yahc_conn_timeline
     yahc_conn_request
     yahc_conn_response
+    yahc_conn_attempt
     yahc_conn_attempts_left
     yahc_conn_socket_cache_id
+    yahc_conn_register_error
 /;
 
 our %EXPORT_TAGS = (all => \@EXPORT_OK);
@@ -258,6 +261,7 @@ sub yahc_conn_errors        { $_[0]->{errors}   }
 sub yahc_conn_timeline      { $_[0]->{timeline} }
 sub yahc_conn_request       { $_[0]->{request}  }
 sub yahc_conn_response      { $_[0]->{response} }
+sub yahc_conn_attempt       { $_[0]->{attempt}  }
 sub yahc_conn_attempts_left { $_[0]->{attempt} > $_[0]->{retries} ? 0 : $_[0]->{retries} - $_[0]->{attempt} + 1 }
 
 sub yahc_conn_target {
@@ -352,7 +356,6 @@ sub _check_stop_condition {
 sub _set_init_state {
     my ($self, $conn_id) = @_;
 
-    my $socket_cache = $self->{socket_cache};
     my $conn = $self->{connections}{$conn_id}  or die "YAHC: unknown connection id $conn_id\n";
 
     $conn->{response} = { status => 0 };
@@ -360,71 +363,71 @@ sub _set_init_state {
     _register_in_timeline($conn, "new state %s", _strstate($conn->{state})) if exists $conn->{debug_or_timeline};
     _call_state_callback($self, $conn, 'init_callback') if exists $conn->{has_init_callback};
 
-    while (1) {
-        _close_or_cache_socket($self, $conn, 1); # force connection close if any (likely not)
-        my $watchers = _delete_watchers_but_lifetime_timer($self, $conn_id); # implicit stop of all watchers
-        return _set_user_action_state($self, $conn_id, YAHC::Error::RETRY_LIMIT(), "retries limit reached")
-            if $conn->{attempt} > $conn->{retries};
+    retry:
+    _close_or_cache_socket($self, $conn, 1); # force connection close if any (likely not)
+    my $watchers = _delete_watchers_but_lifetime_timer($self, $conn_id); # implicit stop of all watchers
 
-        # don't move it before if statement
-        my $attempt = ++$conn->{attempt};
+    return _set_user_action_state($self, $conn_id, YAHC::Error::RETRY_LIMIT(), "retries limit reached")
+        if $conn->{attempt} > $conn->{retries};
 
-        my $request = $conn->{request};
-        _set_request_timer($self, $conn_id)    if $request->{request_timeout};
-        _set_connection_timer($self, $conn_id) if $request->{connect_timeout};
-        _set_drain_timer($self, $conn_id)      if $request->{drain_timeout};
+    # don't move it before if statement
+    my $attempt = $conn->{attempt}++;
+    if ($attempt == 0 || !exists $conn->{request}{_backoff}) {
+        goto retry if _init_helper($self, $conn_id) == 1;
+    } else {
+        my $backoff_delay = eval { $conn->{request}{_backoff}->($conn) };
 
-        my ($host, $ip, $port, $scheme) = eval { _get_next_target($conn) };
         if (my $error = $@) {
-            _register_error($conn, YAHC::Error::CONNECT_ERROR(), "Connection attempt failed: $error");
-            goto schedule_next_attempt;
-        }
-
-        _register_in_timeline($conn, "Target $scheme://$host:$port ($ip:$port) chosen for attempt #$attempt")
-            if exists $conn->{debug_or_timeline};
-
-        my $socket_cache_id; $socket_cache_id = yahc_conn_socket_cache_id($conn) if defined $socket_cache;
-        if (defined $socket_cache_id && exists $socket_cache->{$socket_cache_id}) {
-            _register_in_timeline($conn, "reuse socket '$socket_cache_id'") if $conn->{debug_or_timeline};
-            my $sock = $watchers->{_fh} = delete $socket_cache->{$socket_cache_id};
-            $watchers->{io} = $self->{loop}->io($sock, EV::WRITE, sub {});
-            _set_write_state($self, $conn_id);
-            return;
-        }
-
-        my $sock = eval { _build_socket_and_connect($ip, $port) };
-        if (my $error = $@) {
-            _register_error($conn, YAHC::Error::CONNECT_ERROR(), "Connection attempt failed: $error");
-            goto schedule_next_attempt;
-        };
-
-        _set_connecting_state($self, $conn_id, $sock);
-        return;
-
-        schedule_next_attempt:
-        next unless exists $request->{_backoff};
-
-        $watchers = _delete_watchers_but_lifetime_timer($self, $conn_id); # drop all timers except lifetime
-        return _set_user_action_state($self, $conn_id, YAHC::Error::RETRY_LIMIT(), "retries limit reached")
-            if $conn->{attempt} > $conn->{retries};
-
-        my $backoff_delay = eval { $request->{_backoff}->($conn) };
-        if (my $error = $@) {
-            _set_user_action_state($self, $conn_id, YAHC::Error::CALLBACK_ERROR() | YAHC::Error::TERMINAL_ERROR(),
+            return _set_user_action_state($self, $conn_id, YAHC::Error::CALLBACK_ERROR() | YAHC::Error::TERMINAL_ERROR(),
                 "exception in backoff callback (close connection): $error");
-            return;
         };
 
         $self->{loop}->now_update;
         _register_in_timeline($conn, "setting backoff_timer to %.3fs", $backoff_delay) if exists $conn->{debug_or_timeline};
         $watchers->{backoff_timer} = $self->{loop}->timer($backoff_delay, 0, sub {
             _register_in_timeline($conn, "backoff timer of %.3fs expired", $backoff_delay) if exists $conn->{debug_or_timeline};
-            _set_init_state($self, $conn_id);
+            _set_init_state($self, $conn_id) if _init_helper($self, $conn_id) == 1;
         });
-        return;
     }
 }
 
+sub _init_helper {
+    my ($self, $conn_id) = @_;
+
+    my $conn = $self->{connections}{$conn_id}  or die "YAHC: unknown connection id $conn_id\n";
+    my $watchers = $self->{watchers}{$conn_id} or die "YAHC: no watchers for connection id $conn_id\n";
+
+    my $request = $conn->{request};
+    _set_request_timer($self, $conn_id)    if $request->{request_timeout};
+    _set_connection_timer($self, $conn_id) if $request->{connect_timeout};
+    _set_drain_timer($self, $conn_id)      if $request->{drain_timeout};
+
+    eval {
+        my ($host, $ip, $port, $scheme) = _get_next_target($conn);
+        _register_in_timeline($conn, "Target $scheme://$host:$port ($ip:$port) chosen for attempt #%d", $conn->{attempt})
+            if exists $conn->{debug_or_timeline};
+
+        my $socket_cache = $self->{socket_cache};
+        my $socket_cache_id; $socket_cache_id = yahc_conn_socket_cache_id($conn) if defined $socket_cache;
+        if (defined $socket_cache_id && exists $socket_cache->{$socket_cache_id}) {
+            _register_in_timeline($conn, "reuse socket '$socket_cache_id'") if $conn->{debug_or_timeline};
+            my $sock = $watchers->{_fh} = delete $socket_cache->{$socket_cache_id};
+            $watchers->{io} = $self->{loop}->io($sock, EV::WRITE, sub {});
+            _set_write_state($self, $conn_id);
+        } else {
+            my $sock = _build_socket_and_connect($ip, $port);
+            _set_connecting_state($self, $conn_id, $sock);
+        }
+
+        1;
+    } or do {
+        yahc_conn_register_error($conn, YAHC::Error::CONNECT_ERROR(), "Connection attempt failed: $@");
+        return 1;
+    };
+
+    return 0;
+}
+
 sub _set_connecting_state {
     my ($self, $conn_id, $sock) = @_;
 
@@ -438,14 +441,14 @@ sub _set_connecting_state {
     my $connecting_cb = _get_safe_wrapper($self, $conn, sub {
         my $sockopt = getsockopt($sock, SOL_SOCKET, SO_ERROR);
         if (!$sockopt) {
-            _register_error($conn, YAHC::Error::CONNECT_ERROR(), "Failed to do getsockopt(): $!");
+            yahc_conn_register_error($conn, YAHC::Error::CONNECT_ERROR(), "Failed to do getsockopt(): $!");
             _set_init_state($self, $conn_id);
             return;
         }
 
         if (my $err = unpack("L", $sockopt)) {
             my $strerror = POSIX::strerror($err) || '<unknown POSIX error>';
-            _register_error($conn, YAHC::Error::CONNECT_ERROR(), "Failed to connect: $strerror");
+            yahc_conn_register_error($conn, YAHC::Error::CONNECT_ERROR(), "Failed to connect: $strerror");
             _set_init_state($self, $conn_id);
             return;
         }
@@ -522,7 +525,7 @@ sub _set_ssl_handshake_state {
             return $w->events(EV::WRITE) if $IO::Socket::SSL::SSL_ERROR == SSL_WANT_WRITE;
         }
 
-        _register_error($conn, YAHC::Error::SSL_ERROR(), "Failed to complete SSL handshake: <$!> SSL_ERROR: <$IO::Socket::SSL::SSL_ERROR>");
+        yahc_conn_register_error($conn, YAHC::Error::SSL_ERROR(), "Failed to complete SSL handshake: <$!> SSL_ERROR: <$IO::Socket::SSL::SSL_ERROR>");
         _set_init_state($self, $conn_id);
     });
 
@@ -562,16 +565,16 @@ sub _set_write_state {
                     return $w->events(EV::WRITE) if $IO::Socket::SSL::SSL_ERROR == SSL_WANT_WRITE;
                 }
 
-                _register_error($conn, YAHC::Error::WRITE_ERROR() | YAHC::Error::SSL_ERROR(),
+                yahc_conn_register_error($conn, YAHC::Error::WRITE_ERROR() | YAHC::Error::SSL_ERROR(),
                                 "Failed to send HTTPS data: <$!> SSL_ERROR: <$IO::Socket::SSL::SSL_ERROR>");
                 return _set_init_state($self, $conn_id);
             }
 
             return if $! == EWOULDBLOCK || $! == EINTR || $! == EAGAIN;
-            _register_error($conn, YAHC::Error::WRITE_ERROR(), "Failed to send HTTP data: $!");
+            yahc_conn_register_error($conn, YAHC::Error::WRITE_ERROR(), "Failed to send HTTP data: $!");
             _set_init_state($self, $conn_id);
         } elsif ($wlen == 0) {
-            _register_error($conn, YAHC::Error::WRITE_ERROR(), "syswrite returned 0");
+            yahc_conn_register_error($conn, YAHC::Error::WRITE_ERROR(), "syswrite returned 0");
             _set_init_state($self, $conn_id);
         } else {
             substr($buf, 0, $wlen, '');
@@ -615,19 +618,19 @@ sub _set_read_state {
                     return $w->events(EV::WRITE) if $IO::Socket::SSL::SSL_ERROR == SSL_WANT_WRITE;
                 }
 
-                _register_error($conn, YAHC::Error::READ_ERROR() | YAHC::Error::SSL_ERROR(),
+                yahc_conn_register_error($conn, YAHC::Error::READ_ERROR() | YAHC::Error::SSL_ERROR(),
                                 "Failed to receive HTTPS data: <$!> SSL_ERROR: <$IO::Socket::SSL::SSL_ERROR>");
                 return _set_init_state($self, $conn_id);
             }
 
             return if $! == EWOULDBLOCK || $! == EINTR || $! == EAGAIN;
-            _register_error($conn, YAHC::Error::READ_ERROR(), "Failed to receive HTTP data: $!");
+            yahc_conn_register_error($conn, YAHC::Error::READ_ERROR(), "Failed to receive HTTP data: $!");
             _set_init_state($self, $conn_id);
         } elsif ($rlen == 0) {
             if ($content_length > 0) {
-                _register_error($conn, YAHC::Error::READ_ERROR(), "Premature EOF, expect %d bytes more", $content_length - length($buf));
+                yahc_conn_register_error($conn, YAHC::Error::READ_ERROR(), "Premature EOF, expect %d bytes more", $content_length - length($buf));
             } else {
-                _register_error($conn, YAHC::Error::READ_ERROR(), "Premature EOF");
+                yahc_conn_register_error($conn, YAHC::Error::READ_ERROR(), "Premature EOF");
             }
             _set_init_state($self, $conn_id);
         } else {
@@ -710,7 +713,7 @@ sub _set_user_action_state {
 
     $conn->{state} = YAHC::State::USER_ACTION();
     _register_in_timeline($conn, "new state %s", _strstate($conn->{state})) if exists $conn->{debug_or_timeline};
-    _register_error($conn, $error, $strerror) if $error != YAHC::Error::NO_ERROR;
+    yahc_conn_register_error($conn, $error, $strerror) if $error != YAHC::Error::NO_ERROR;
 
     _close_or_cache_socket($self, $conn, $error != YAHC::Error::NO_ERROR);
     return _set_completed_state($self, $conn_id) unless exists $conn->{has_callback};
@@ -722,7 +725,7 @@ sub _set_user_action_state {
         1;
     } or do {
         my $error = $@ || 'zombie error';
-        _register_error($conn, YAHC::Error::CALLBACK_ERROR(), "Exception in user action callback (close connection): $error");
+        yahc_conn_register_error($conn, YAHC::Error::CALLBACK_ERROR() | YAHC::Error::TERMINAL_ERROR(), "Exception in user action callback (close connection): $error");
         $self->{state} = YAHC::State::COMPLETED();
     };
 
@@ -730,7 +733,7 @@ sub _set_user_action_state {
 
     my $state = $conn->{state};
     if (yahc_terminal_error($error)) {
-        _register_error($conn, YAHC::Error::CALLBACK_ERROR(), "ignoring changed state due to terminal error")
+        yahc_conn_register_error($conn, YAHC::Error::CALLBACK_ERROR() | YAHC::Error::TERMINAL_ERROR(), "ignoring changed state due to terminal error")
             unless $state == YAHC::State::USER_ACTION() || $state == YAHC::State::COMPLETED();
         _set_completed_state($self, $conn_id, 1);
         return
@@ -743,7 +746,7 @@ sub _set_user_action_state {
     } elsif ($state == YAHC::State::USER_ACTION() || $state == YAHC::State::COMPLETED()) {
         _set_completed_state($self, $conn_id);
     } else {
-        _register_error($conn, YAHC::Error::CALLBACK_ERROR(), "callback set unsupported state");
+        yahc_conn_register_error($conn, YAHC::Error::CALLBACK_ERROR() | YAHC::Error::TERMINAL_ERROR(), "callback set unsupported state");
         _set_completed_state($self, $conn_id);
     }
 }
@@ -849,9 +852,9 @@ sub yahc_conn_socket_cache_id {
 # Timers
 ################################################################################
 
-sub _set_request_timer    { $_[0]->_set_until_state_timer($_[1], 'request_timeout', YAHC::State::USER_ACTION(), YAHC::Error::REQUEST_TIMEOUT()) }
-sub _set_connection_timer { $_[0]->_set_until_state_timer($_[1], 'connect_timeout', YAHC::State::CONNECTED(),   YAHC::Error::CONNECT_TIMEOUT()) }
-sub _set_drain_timer      { $_[0]->_set_until_state_timer($_[1], 'drain_timeout',   YAHC::State::READING(),     YAHC::Error::DRAIN_TIMEOUT())   }
+sub _set_request_timer    { $_[0]->_set_until_state_timer($_[1], 'request_timeout', YAHC::State::USER_ACTION(), YAHC::Error::TIMEOUT() | YAHC::Error::REQUEST_TIMEOUT()) }
+sub _set_connection_timer { $_[0]->_set_until_state_timer($_[1], 'connect_timeout', YAHC::State::CONNECTED(),   YAHC::Error::TIMEOUT() | YAHC::Error::CONNECT_TIMEOUT()) }
+sub _set_drain_timer      { $_[0]->_set_until_state_timer($_[1], 'drain_timeout',   YAHC::State::READING(),     YAHC::Error::TIMEOUT() | YAHC::Error::DRAIN_TIMEOUT())   }
 
 sub _set_until_state_timer {
     my ($self, $conn_id, $timeout_name, $state, $error_to_report) = @_;
@@ -866,7 +869,7 @@ sub _set_until_state_timer {
 
     my $timer_cb = sub { # there is nothing what can throw exception
         if ($conn->{state} < $state) {
-            _register_error($conn, $error_to_report, "$timeout_name of %.3fs expired", $timeout);
+            yahc_conn_register_error($conn, $error_to_report, "$timeout_name of %.3fs expired", $timeout);
             _set_init_state($self, $conn_id);
         } else {
             _register_in_timeline($conn, "delete $timer_name") if exists $conn->{debug_or_timeline};
@@ -895,8 +898,8 @@ sub _set_lifetime_timer {
 
     $self->{loop}->now_update;
     my $w = $watchers->{lifetime_timer} = $self->{loop}->timer_ns($timeout, 0, sub {
-        _set_user_action_state($self, $conn_id, YAHC::Error::LIFETIME_TIMEOUT() | YAHC::Error::TERMINAL_ERROR(),
-            sprintf("lifetime timer of %.3fs expired", $timeout)) if $conn->{state} < YAHC::State::COMPLETED();
+        _set_user_action_state($self, $conn_id, YAHC::Error::TIMEOUT() | YAHC::Error::LIFETIME_TIMEOUT() | YAHC::Error::TERMINAL_ERROR(),
+            sprintf("lifetime_timeout of %.3fs expired", $timeout)) if $conn->{state} < YAHC::State::COMPLETED();
     });
 
     $w->priority(2); # set highest priority
@@ -998,15 +1001,17 @@ sub _call_state_callback {
     my $cb = $self->{callbacks}{$conn->{id}}{$cb_name};
     return unless $cb;
 
+    _register_in_timeline($conn, "calling $cb_name callback") if exists $conn->{debug_or_timeline};
+
     eval {
         $cb->($conn);
         1;
     } or do {
         my $error = $@ || 'zombie error';
-        _register_error($conn, YAHC::Error::CALLBACK_ERROR(), "exception in state callback (ignore error): $error");
+        yahc_conn_register_error($conn, YAHC::Error::CALLBACK_ERROR(), "exception in state callback (ignore error): $error");
     };
 
-    # $self->{loop}->now_update; # XXX expect state callbacks to be small
+    $self->{loop}->now_update;
 }
 
 sub _get_safe_wrapper {
@@ -1028,11 +1033,11 @@ sub _register_in_timeline {
     push @{ $conn->{timeline} ||= [] }, [ $event, $conn->{state}, Time::HiRes::time ] if exists $conn->{keep_timeline};
 }
 
-sub _register_error {
+sub yahc_conn_register_error {
     my ($conn, $error, $format, @arguments) = @_;
     my $strerror = sprintf("$format", @arguments);
     _register_in_timeline($conn, "strerror='$strerror' error=$error") if exists $conn->{debug_or_timeline};
-    push @{ $conn->{errors} ||= [] }, [ $error, $strerror, [ @{ $conn->{selected_target} } ], Time::HiRes::time ];
+    push @{ $conn->{errors} ||= [] }, [ $error, $strerror, [ @{ $conn->{selected_target} } ], Time::HiRes::time, $conn->{attempt} ];
 }
 
 sub _strstate {
@@ -1094,13 +1099,13 @@ YAHC is fast & minimal low-level asynchronous HTTP client intended to be used
 where you control both the client and the server. Is especially suits cases
 where set of requests need to be executed against group of machines.
 
-It is C<NOT> a general HTTP user agent, it doesn't support redirects,
+It is B<NOT> a general HTTP user agent, it doesn't support redirects,
 proxies and any number of other advanced HTTP features like (in
 roughly descending order of feature completeness) L<LWP::UserAgent>,
 L<WWW::Curl>, L<HTTP::Tiny>, L<HTTP::Lite> or L<Furl>. This library is
 basically one step above manually talking HTTP over sockets.
 
-YAHC supports SSL and socket reuse (later is in experimental mode).
+YAHC supports SSL and socket reuse (latter is in experimental mode).
 
 =head1 STATE MACHINE
 
@@ -1147,13 +1152,13 @@ There are three paths of workflow:
 
 In normal situation a connection after being initialized goes through state:
 
-- RESOLVE DNS
+- RESOLVE DNS (not implemented)
 
 - CONNECTING - wait finishing of handshake
 
 - CONNECTED
 
-- WRITTING - sending request body
+- WRITING - sending request body
 
 - READING - awaiting and reading response
 
@@ -1167,7 +1172,7 @@ SSL connection has extra state SSL_HANDSHAKE after CONNECTED state. State
 =item 2) Retry path (right line).
 
 In case of IO error during normal execution YAHC retries connection
-C<retries> times. In practise this means that connection goes back to
+C<retries> times. In practice this means that connection goes back to
 INITIALIZED state.
 
 =item 3) Failure path (left line).
@@ -1181,7 +1186,7 @@ ACTION' (see below).
 
 'USER ACTION' state is called right before connection if going to enter
 'COMPLETED' state (with either failed or successful results) and is meant
-to give a change to user to interupt the workflow.
+to give a chance to user to interrupt the workflow.
 
 'USER ACTION' state is entered in these circumstances:
 
@@ -1242,7 +1247,7 @@ C<$error> having YAHC::Error::TERMINAL_ERROR() bit set. One can use
 C<yahc_terminal_error> helper to detect such case.
 
 Note that C<callback> should NOT throw exception. If so the connection will be
-imidiately closed.
+immediately closed.
 
 =head1 METHODS
 
@@ -1263,7 +1268,7 @@ $yahc_storage object it's fine to use YAHC object inside request callback:
     });
 
 However, user has to guarantee that both $yahc and $yahc_storage objects are
-kept in the same namespace. So, they will be destroyed at the same time.
+kept in the same scope. So, they will be destroyed at the same time.
 
 C<new> can be passed with all parameters supported by C<request>. They
 will be inherited by all requests.
@@ -1280,12 +1285,12 @@ HashRef.
     my ($yahc, $yahc_storage) = YAHC->new({ socket_cache => {} });
 
 In this case YAHC maintains unused sockets keyed on C<join($;, $$, $host,
-$port, $scheme)>. We use $; so we can use the $socket_cache->{$$, $host, $port,
-$scheme} idiom to access the cache.
+$port, $scheme)>. We use C<$;> so we can use the C<< $socket_cache->{$$, $host, $port,
+$scheme} >> idiom to access the cache.
 
 It's up to user to control the cache. It's also up to user to set necessary
-request headers for keep-alive. YAHC does not cache socket in cases of a error,
-HTTP/1.0 and when server explicitly instruct to close connection (i.e header
+request headers for keep-alive. YAHC does not cache socket in cases of an error,
+HTTP/1.0 and when server explicitly instructs to close connection (i.e. header
 'Connection' = 'close').
 
 =head3 account_for_signals
@@ -1295,7 +1300,7 @@ why:
 
 =over 4
 
-exerpt from EV documentation http://search.cpan.org/~mlehmann/EV-4.22/EV.pm#PERL_SIGNALS
+excerpt from EV documentation L<http://search.cpan.org/~mlehmann/EV-4.22/EV.pm#PERL_SIGNALS>
 
 While Perl signal handling (%SIG) is not affected by EV, the behaviour with EV
 is as the same as any other C library: Perl-signals will only be handled when
@@ -1305,7 +1310,7 @@ an event callback is invoked.
 =back
 
 In practise this means that none of set %SIG handlers will be called until EV
-calls one of perl callbacks. Which, in some cases, may take long time. By
+calls one of perl callbacks. Which, in some cases, may take a long time. By
 setting C<account_for_signals> YAHC adds C<EV::check> watcher with empty
 callback effectively making EV calling the callback on every iteration. The
 trickery comes at some performance cost. This is what EV documentation says
@@ -1445,7 +1450,6 @@ third attempts will be delay by one second each.
         backoff_delay => 1,
     });
 
-
 C<backoff_delay> can be set in two ways:
 
 =over 4
@@ -1463,7 +1467,7 @@ The default value is C<undef>, meaning no delay.
 =head3 callbacks
 
 The value of C<init_callback>, C<connecting_callback>, C<connected_callback>,
-C<writing_callback>, C<reading_callback> is CodeRef to a subroutine which is
+C<writing_callback>, C<reading_callback> is a reference to a subroutine which is
 called upon reaching corresponding state. Any exception thrown in the
 subroutine will be ignored.
 
@@ -1474,7 +1478,7 @@ Also see L<LIMITATIONS>
 
 =head3 ssl_options
 
-Performing HTTPS request the value of ssl_options extended by two parameters
+Performing HTTPS requires the value of C<ssl_options> extended by two parameters
 set to current hostname:
 
         SSL_verifycn_name => $hostname,
@@ -1573,7 +1577,7 @@ connection is in 'USER ACTION' state.
 Retries given connection. C<yahc_retry_conn> should be called only if
 C<yahc_conn_attempts_left> returns positive value. Otherwise, it exits silently.
 
-Intended usege is to retry transient failures or try different host:
+Intended usage is to retry transient failures or to try different host:
 
     use YAHC qw/
         yahc_retry_conn
@@ -1596,7 +1600,7 @@ Intended usege is to retry transient failures or try different host:
 
     $yahc->run;
 
-C<yahc_conn_attempts_left> is meant to be called inside C<callback> similarly
+C<yahc_retry_conn> is meant to be called inside C<callback> similarly
 to C<yahc_reinit_conn>.
 
 =head2 yahc_conn_id
@@ -1605,7 +1609,7 @@ Return id of given connection.
 
 =head2 yahc_conn_state
 
-Retrun state of given connection
+Return state of given connection.
 
 =head2 yahc_conn_target
 
@@ -1629,9 +1633,40 @@ items:
     error string
     ArrayRef of host, ip, port, scheme
     time when the error happened
+    attempt when the error happened
 
 =back
 
+=head2 yahc_conn_register_error
+
+C<yahc_conn_register_error> adds new record in connection's error list. This
+functions is used internally for keeping track of all low-level errors during
+connection's lifetime. It can be also used by users for high-level errors such
+as 50x responses. The function takes C<$conn>, C<$error> which is one of
+C<YAHC::Error> constants and error description. Error description can be passed
+in sprintf manner. For example:
+
+    $yahc->request({
+        ...
+        callback => sub {
+            ...
+            my $conn = $_[0];
+            my $status = $conn->{response}{status} || 0;
+            if ($status == 503 || $status == 504) {
+                yahc_conn_register_error(
+                    $conn,
+                    YAHC::Error::RESPONSE_ERROR(),
+                    "server returned %d",
+                    $status
+                );
+
+                yahc_retry_conn($conn);
+                return;
+            }
+            ...
+        }
+    });
+
 =head2 yahc_conn_last_error
 
 Return last error appeared in connection. See C<yahc_conn_errors>.
@@ -1654,6 +1689,102 @@ Return request of given connection. See C<request>.
 
 Return response of given connection. See C<request>.
 
+=head2 yahc_conn_attempt
+
+Return number of current attempt starting from 0.
+
+=head2 yahc_conn_attempts_left
+
+Return number of attempts left.
+
+=head2 yahc_conn_socket_cache_id
+
+Return socket_cache id for given connection. Should be used to generate key for
+C<socket_cache>. If connection is not initialized yet C<undef> is returned.
+
+=head1 ERRORS
+
+YAHC provides set of constants for errors. Each constant returns bitmask which
+can be used to detect presence of a particular error, for example, in
+C<callback>. There is one exception: YAHC::Error::NO_ERROR() return 0
+indicating no error during request execution.
+
+Error handling code can look like following:
+
+    $yahc->request({
+        ...
+        callback => sub {
+            my (
+                $conn,          # connection 'object'
+                $error,         # one of YAHC::Error::* constants
+                $strerror       # string representation of error
+            ) = @_;
+
+            if ($error & YAHC::Error::TIMEOUT()) {
+                # A timeout has happend. Use one of YAHC::Error::*_TIMEOUT()
+                # constants for more clarification
+            } elsif ($error & YAHC::Error::SSL_ERROR()) {
+                # We had some issues with SSL. $error might have
+                # YAHC::Error::READ_ERROR() or YAHC::Error::WRITE_ERROR()
+                # indicating whether is was read or write error.
+            } elsif (...) { # etc
+            }
+        }
+    });
+
+The list of error constants. The names are self-explanatory in many cases:
+
+=over 4
+
+=item C<YAHC::Error::NO_ERROR()>
+
+Return value 0 (not a bitmask)> meaning no error
+
+=item C<YAHC::Error::REQUEST_TIMEOUT()>
+
+=item C<YAHC::Error::CONNECT_TIMEOUT()>
+
+=item C<YAHC::Error::DRAIN_TIMEOUT()>
+
+=item C<YAHC::Error::LIFETIME_TIMEOUT()>
+
+=item C<YAHC::Error::TIMEOUT()>
+
+=item C<YAHC::Error::RETRY_LIMIT()>
+
+The connection has exhausted all available retries. This error is usually
+returned to C<callback>. Check connection's errors via C<yahc_conn_errors> to
+inspect the reasons of failures for each individual attempt.
+
+=item C<YAHC::Error::CONNECT_ERROR()>
+
+=item C<YAHC::Error::READ_ERROR()>
+
+=item C<YAHC::Error::WRITE_ERROR()>
+
+=item C<YAHC::Error::SSL_ERROR()>
+
+=item C<YAHC::Error::REQUEST_ERROR()>
+
+not used
+
+=item C<YAHC::Error::RESPONSE_ERROR()>
+
+Server returned unparsable response
+
+=item C<YAHC::Error::CALLBACK_ERROR()>
+
+Usually represents exception in one of the callbacks
+
+=item C<YAHC::Error::TERMINAL_ERROR()>
+
+This bit is set when connection cannot be retried anymore and is forced to
+complete
+
+=item C<YAHC::Error::INTERNAL_ERROR()>
+
+=back
+
 =head1 REPOSITORY
 
 L<https://github.com/ikruglov/YAHC>
@@ -1682,7 +1813,7 @@ to make sure that *all* data passed to YAHC is unflagged binary strings.
 
 =item * State 'RESOLVE DNS' is not implemented yet.
 
-=item * YAHC currently don't support servers returning a http body without an
+=item * YAHC currently doesn't support servers returning a http body without an
 accompanying C<Content-Length> header; bodies B<MUST> have a C<Content-Length>
 or we won't pick them up.
 
